import shutil
import html.parser

from gdrive_util import GDrive

from markdown_parser import MarkdownParser
import mistune
from whoosh.fields import *
import whoosh.index as index
import os
import os.path
import codecs
from whoosh.qparser import MultifieldParser, QueryParser
from whoosh.analysis import StemmingAnalyzer


"""
cheeseburger_search.py 


Auth:

- this program uses google drive and therefore uses
  google's oauth mechanism to authenticate.
- gdrive_util.py takes care of creating the API instance
  from keys and creating an API service.



Flow:

program will:
   - create a Search object
   - call add_all_documents
   - calls open_index if no index exists
   - grab a list of repos
   - walk every issue
        - each issue body is (by itself) a document
        - each comment body is also a document
   - call add_document on each issue and each comment
   - add_document parses Markdown/adds docs to index following schema

program will occasionally:
   - ask to update the index
   - this opens the index, and adds all files in the path to a list
   - it then finds files that were deleted and files that were changed
   - it passes these to the indexer, then it's done

program calls the search function:
    - takes query from user
    - calls searcher.search(query, ...)
    - parses results from searcher into user-friendly search results
"""


class SearchResult:
    score = 1.0
    path = None
    content = ""
    content_highlight = ""
    headlines = None
    tags = ""


class DontEscapeHtmlInCodeRenderer(mistune.Renderer):
    def __init__(self, **kwargs):
        super(DontEscapeHtmlInCodeRenderer, self).__init__(**kwargs)

    def block_code(self, code, lang):
        if not lang:
            return '<pre><code>%s\n</code></pre>\n' % code
        return '<pre><code class="lang-%s">%s\n</code></pre>\n' % (lang, code)

    def codespan(self, text):
        return '<code>%s</code>' % text.rstrip()


class Search:
    ix = None
    index_folder = None
    markdown = mistune.Markdown(renderer=DontEscapeHtmlInCodeRenderer(), escape=False)
    html_parser = html.parser.HTMLParser()
    schema = None

    def __init__(self, index_folder):
        self.open_index(index_folder)

    def open_index(self, index_folder, create_new=False):
        """
        Create a schema,
        and create/open a search index
        that lives on disk.
        """
        self.index_folder = index_folder
        if create_new:
            if os.path.exists(index_folder):
                shutil.rmtree(index_folder)
                print("deleted index folder: " + index_folder)

        if not os.path.exists(index_folder):
            os.mkdir(index_folder)

        exists = index.exists_in(index_folder)
        stemming_analyzer = StemmingAnalyzer()

        # cheeseburger schema
        # useful:
        # https://developers.google.com/drive/api/v3/reference/files
        #
        # id -> id
        # url-> webViewLink
        # timestamp -> createdTime
        # owner -> owners[]['emailAddress']
        # owner_name -> owners[]['displayName']
        # title -> name
        # content -> (from pandoc)

        # IMPORTANT:
        # This is where the schema is defined.

        schema = Schema(
                id=ID(stored=True,unique=True),
                url=ID(stored=True, unique=True),
                mimetype=ID(stored=True),
                timestamp=ID(stored=True),
                owner_email=ID(stored=True),
                owner_name=ID(stored=True),
                title=TEXT(stored=True),
                content=TEXT(stored=True, analyzer=stemming_analyzer)
        )

        # Now that we have a schema,
        # make an index!
        if not exists:
            self.ix = index.create_in(index_folder, schema)
        else:
            self.ix = index.open_dir(index_folder)


    def add_all_documents(self, credentials_file, collection, config, create_new_index=False):
        """
        Add all issues in a given github repo to the search index.

        Replaces add_all_files

        Takes as inputs:
        - github access token
        - list of github repos
        - github org/user owning these repos
        - location of the whoosh config file for configuring the search engine
        """

        if create_new_index:
            self.open_index(self.index_folder, create_new=True)

        gd = GDrive()
        service = gd.get_service()

        # Iindex each document:
        # 
        # Step 1: walk each doc in google drive. 
        # Step 2: index it.
        # Step 3: grab a beer.
        #

        # -----
        # Set of all documents on Google Drive:

        # Call the Drive v3 API

        results = service.files().list(
            pageSize=100, fields="nextPageToken, files(id, kind, mimeType, name, owners, webViewLink, createdTime)").execute()

        items = results.get('files', [])

        # To use the next token, 
        # just say results['nextPageToken']
        # otherwise use items for the files
        # 
        # TODO:
        # Tapping out at 100, use nextPageToken to get all later
        for item in items:
            drive_ids.add(item['id'])

        count = 0
        for item in items:

            # If we have already indexed this document, 
            # drop the old record first
            if item['id'] in indexed_ids:
                writer.delete_by_term('id',item['id'])

            # IMPORTANT:
            # This is where the search documents are actually created.

            mimetype = re.split('[/\.]', item['mimeType'])[-1]
            writer.add_document(
                    id = item['id'],
                    url = item['webViewLink'],
                    mimetype = mimetype,
                    timestamp = item['createdTime'],
                    owner_email = item['owners'][0]['emailAddress'],
                    owner_name = item['owners'][0]['displayName'],
                    title = item['name']
            )
            count += 1

        # TODO:
        # Major todo item:
        # - download items using id and get endpoint: https://developers.google.com/drive/api/v3/reference/files/get
        # - convert to markdown
        # - add content to content field

        writer.commit()
        print("Done, created %d documents in the index" % count)



    def update_index_incremental(self, credentials_file, collection, config, create_new_index=False):
        """
        Update the index of issues of a given github repo.

        Takes as inputs:
        - github access token
        - list of github repos
        - github org/user owning these repos
        - location of the whoosh config file for configuring the search engine
        """

        if create_new_index:
            self.open_index(self.index_folder, create_new=True)

        gd = GDrive()
        service = gd.get_service()

        # PoC||GTFO

        # Re-index each document:
        # 
        # Step 1: build list of indexed documents
        # Step 2: build list of documents on google drive
        # Step 3: if indexed documents not on google drive, delete
        # Step 4: if indexed documents on google drive, delete and reindex
        # Step 5: if non-indexed documents on google drive, index
        #
        # TODO:
        # Can make Step 4 shorter by storing hash of contents.

        # -----
        # Set of all indexed documents:
        indexed_ids = set()

        with self.ix.searcher() as searcher:
            writer = self.ix.writer()

            # Loop over the stored fields in the index
            # (i.e., each record)
            for fields in searcher.all_stored_fields():
                indexed_id = fields['id']
                indexed_ids.add(indexed_id)

        # -----
        # Set of all documents on Google Drive:
        drive_ids = set()

        # Call the Drive v3 API

        ## short record
        #results = service.files().list(
        #    pageSize=100, fields="nextPageToken, files(id, name)").execute()

        # long record
        results = service.files().list(
            pageSize=100, fields="nextPageToken, files(id, kind, mimeType, name, owners, webViewLink, createdTime)").execute()

        items = results.get('files', [])

        # To use the next token, 
        # just say results['nextPageToken']
        # otherwise use items for the files
        # 
        # TODO:
        # Tapping out at 100, use nextPageToken to get all later
        for item in items:
            drive_ids.add(item['id'])

        # Remove documents in the index that are not in this drive
        for indexed_id in indexed_ids:
            if indexed_id not in drive_ids:
                writer.delete_by_term('id',indexed_id)

        count = 0

        for item in items:

            # If we have already indexed this document, 
            # drop the old record first
            if item['id'] in indexed_ids:
                writer.delete_by_term('id',item['id'])

            # IMPORTANT:
            # This is where the search documents are actually created.
            
            mimetype = re.split('[/\.]', item['mimeType'])[-1]
            writer.add_document(
                    id = item['id'],
                    url = item['webViewLink'],
                    timestamp = item['createdTime'],
                    mimetype = mimetype,
                    owner_email = item['owners'][0]['emailAddress'],
                    owner_name = item['owners'][0]['displayName'],
                    title = item['name']
            )
            count += 1

        # TODO:
        # Major todo item:
        # - download items using id and get endpoint: https://developers.google.com/drive/api/v3/reference/files/get
        # - convert to markdown
        # - add content to content field

        writer.commit()
        print("Done, updated %d documents in the index" % count)



    def create_search_result(self, results):
        # Allow larger fragments
        results.fragmenter.maxchars = 300

        # Show more context before and after
        results.fragmenter.surround = 50

        search_results = []
        for r in results:

            # Note: this is where we package things up 
            # for the Jinja template "search.html".
            # For example, the Jinja template
            # contains a {% for e in entries %}
            # and then an {{e.score}}

            sr = SearchResult()
            sr.score = r.score


            # IMPORTANT:
            # update search.html with what you want to see
            # in each search result's "metadata" (links,
            # parent repos, users, etc.)

            # sr variables are available in Jinja
            # r variables are from documents (follow schema)

            sr.id = r['id']
            sr.url = r['url']
            sr.title = r['title']
            sr.mimetype = r['mimetype']

            sr.owner_email = r['owner_email']
            sr.owner_name = r['owner_name']

            #sr.content = r['content']

            #highlights = r.highlights('content')
            #if not highlights:
            #    # just use the first 1,000 words of the document
            #    highlights = self.cap(r['content'], 1000)

            #highlights = self.html_parser.unescape(highlights)
            #html = self.markdown(highlights)
            #sr.content_highlight = html
            sr.content_highlight = '<p>Hello world</p>'

            search_results.append(sr)

        return search_results

    def cap(self, s, l):
        return s if len(s) <= l else s[0:l - 3] + '...'

    def search(self, query_list, fields=None):
        with self.ix.searcher() as searcher:
            query_string = " ".join(query_list)
            query = None
            if "\"" in query_string or ":" in query_string:
                query = QueryParser("content", self.schema).parse(query_string)
            elif len(fields) == 1 and fields[0] == "filename":
                pass
            elif len(fields) == 2:
                pass
            else:
                fields = ['id','url','timestamp','owner_email','owner_name','title','content'] 
            if not query:
                query = MultifieldParser(fields, schema=self.ix.schema).parse(query_string)
            parsed_query = "%s" % query
            print("query: %s" % parsed_query)
            results = searcher.search(query, terms=False, scored=True, groupedby="url")
            search_result = self.create_search_result(results)

        return parsed_query, search_result

    def get_document_total_count(self):
        return self.ix.searcher().doc_count_all()

if __name__ == "__main__":
    search = Search("search_index")
    search.add_all_documents("/Users/charles/codes/cheeseburger-search/config.py")

